<'
# Copyright (c) 2012-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar  7 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include "<"$comp">_control_task.h"
#include "<"$comp">_activity.h"


/* --- local data ---------------------------------------------------------- */


/* <"[--- genom_${comp}_activity_alloc ----------------------------------]"> */

int
genom_<"$comp">_activity_alloc(struct genom_<"$comp">_cntrlstr *cids)
{
  int id;
  int i;

  /* look for the first free activity */
  id = -1;
  for(i = 1; i<=<"$COMP">_MAX_ACTIVITIES; i++) {
    id = (cids->activities.last + i) % <"$COMP">_MAX_ACTIVITIES;
    if (cids->activities.i[id].any.status == ACT_VOID) break;
  }
  if (id < 0) return -1;

  cids->activities.last = id;
  return id;
}


/* <"[--- genom_${comp}_activity_free -----------------------------------]"> */

void
genom_<"$comp">_activity_free(struct genom_<"$comp">_cntrlstr *cids, int id)
{
  cids->activities.i[id].any.status = ACT_VOID;
}


/* === Send final replies ================================================== */

<'foreach s [$component services] {'>
/* <"[$s name]"> */
static __inline__ void
genom_<"$comp">_<"[$s name]">_activity_report(
  struct genom_<"$comp">_cntrlstr *cids,
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
  STATUS s;
  int i;

  /* send final reply */
  assert(a->h.status == ACT_ETHER || a->h.status == ACT_STOPPED);

  if (a->h.status == ACT_STOPPED && a->h.codel == <"$comp">_ether) {
    genom_interrupted_detail d;
    strncpy(d.by, a->h.interruptedby, sizeof(d.by));
    d.by[sizeof(d.by)-1] = *"";
    genom_interrupted(&d);
    a->h.exdetail = (void *)genom_thrown(&a->h.codel);
  }
  if (a->h.codel == <"$comp">_ether)
    a->h.codel = genom_ok;

  if (!a->h.codel) {
    /* success: update after/before array */
    cids->run_map[<"$COMP">_<"[$s name]">_RQSTID] = 1;

    s = csServReplySend(
      cids->csserv, a->h.rid, FINAL_REPLY, OK,
      (char *)&a->out, 0, genom_<"$comp">_<"[$s name]">_encode);
  } else
    s = csServReplySend(
      cids->csserv, a->h.rid, FINAL_REPLY, ERROR, (char *)&a->h, 0,
      genom_<"$comp">_activity_encodex);
  if (s == ERROR) {
    genom_log_warn(1, "could not send output for service %s", "<"[$s name]">");
    s = csServReplySend(
      cids->csserv, a->h.rid, FINAL_REPLY, ERROR, NULL, 0,
      genom_<"$comp">_genom_serialization_encodex);
    if (s == ERROR) {
      genom_log_warn(1, "discarding output for service %s", "<"[$s name]">");
    } else
      genom_log_warn(0, "invalid output for service %s", "<"[$s name]">");
  }

  /* update state port */
  {
    genom_state_component *state = &cids->ports.genom_state.h.buffer;

    if (a->h.stateid >= 0) {
      state->activity._length--;
      memcpy(&state->activity._buffer[a->h.stateid],
             &state->activity._buffer[a->h.stateid+1],
             (state->activity._length - a->h.stateid) *
             sizeof(state->activity._buffer[a->h.stateid]));
      for(i=0; i<<"$COMP">_MAX_ACTIVITIES; i++) {
        if (cids->activities.i[i].any.stateid > a->h.stateid)
          cids->activities.i[i].any.stateid--;
      }
    }
  }

  /* cleanup */
  genom_tfini_<"$comp">_<"[$s name]">_activity(a);
  genom_<"$comp">_activity_free(cids, a->h.aid);
}

<'}'>

/* <"[--- genom_${comp}_activity_report ---------------------------------]"> */

void
genom_<"$comp">_activity_report(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
  switch(a->sid) {
<'foreach s [$component services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      genom_<"$comp">_<"[$s name]">_activity_report(
        cids, (struct genom_<"$comp">_<"[$s name]">_activity *)a);
      return;
<'}'>
  }

  assert(!"code not reached");
}


/* --- genom_<"$comp">_activity_encodex ------------------------------------ */

int
genom_<"$comp">_activity_encodex(char *buffer, int size, char *dst, int maxsize)
{
  struct genom_activity *a = (struct genom_activity *)buffer;

  if (a->codel == genom_unkex_id)
    return genom_<"$comp">_genom_unkex_encodex(a->exdetail, size, dst, maxsize);
  else if (a->codel == genom_syserr_id)
    return genom_<"$comp">_genom_syserr_encodex(a->exdetail, size, dst, maxsize);
<'foreach e [$component throws] {'>
  else if (a->codel == <"[$e cname]">_id)
    return genom_<"$comp">_<"[$e cname]">_encodex(
      a->exdetail, size, dst, maxsize);
<'}'>

  assert(!"unknown genom exception");
  return 0;
}


/* === Interrupt incompatible activities =================================== */

<'foreach s [$component services] {'>
/* <"[$s name]"> */
static __inline__ int
genom_<"$comp">_<"[$s name]">_interrupt_reqd(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
<'  if {[llength [$s interrupt]]} {'>
  struct genom_activity *c;
  int delay;
  int id;

  delay = 0;
  for(id = 0; id<<"$COMP">_MAX_ACTIVITIES; id++) {
    c = &cids->activities.i[id].any;
    if (c->aid == a->aid) continue;
    if (c->status == ACT_VOID) continue;

    switch(c->sid) {
<'    foreach c [$s interrupt] {'>
      case <"$COMP">_<"[$c name]">_RQSTID:
<'    }'>
        switch(c->status) {
          case ACT_ETHER:
          case ACT_STOPPED:
            break;

          case ACT_STOP:
          case ACT_STOPPING:
            delay = 1;
            break;

          case ACT_RUNNING:
            c->status = ACT_STOP;
            c->interruptedby = "<"[$s name]">";
            genom_<"$comp">_schedule_exec(cids);
            delay = 1;
            break;

          case ACT_START:
            c->status = ACT_STOPPED;
            c->codel = <"$comp">_ether;
            genom_<"$comp">_schedule_cntrl(cids);
            break;

          case ACT_VOID: assert(!"code not reached");
        }
        break;
    }
  }
  return delay;
<'  } else {'>
  return 0;
<'  }'>
}

<'}'>

/* <"[--- genom_${comp}_interrupt_reqd ----------------------------------]"> */

int
genom_<"$comp">_interrupt_reqd(
  struct genom_<"$comp">_cntrlstr *cids, struct genom_activity *a)
{
  /* interrupt incompatible activities */
  switch(a->sid) {
<'foreach s [$component services] {'>
    case <"$COMP">_<"[$s name]">_RQSTID:
      return genom_<"$comp">_<"[$s name]">_interrupt_reqd(cids, a);
<'}'>
  }

  assert(!"code not reached");
  return 0;
}
